/*
 * Copyright (c) 2016 Chris Johns
 * All rights resevered
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.org/license/LICENSE.
 */

#ifndef _RTEMS_TINYSSHD_H
#define _RTEMS_TINYSSHD_H

#include <rtems.h>
#include <rtems/shell.h>

#ifdef __cplusplus
extern "C" {
#endif

bool rtems_tinysshd_login_check(
  const char *user,
  const char *passphrase
);

/**
 * @brief SSH command type.
 */
typedef void (*rtems_tinysshd_command)(
  char * /* device name */,
  void * /* arg */
);

/**
 * @brief Tinysshd configuration structure.
 */
typedef struct {
  /**
   * The maximum number of connections. If 0 the default of 4 is used.
   */
  int connections;

  /**
   * Port for the SSHD to listen on,
   */
  int port;

  /**
   * The back is the number of connections the listener can handle. If 0 the
   * default of 1 is used.
   */
  int backlog;

  /**
   * @brief Function invoked for each SSH connection.
   *
   * The first parameter contains the device name.  The second parameter
   * contains the argument pointer of this configuration table.
   */
  rtems_tinysshd_command command;

  /**
   * @brief Argument for command function.
   */
  void* arg;

  /**
   * @brief SSH Task priority.
   *
   * This parameter must be set. It is a Classic priority level.
   */
  rtems_task_priority priority;

  /**
   * @brief SSH Task stack size.
   */
  size_t stack_size;

  /**
   * @brief Login check function.
   *
   * Method used for login checks.  Use @c NULL to disable a login check.
   */
  rtems_shell_login_check_t login_check;

  /**
   * @brief Keep standard IO of the caller.
   *
   * SSHD takes over the standard input, output and error associated with
   * task, if this parameter is set to @c true.  In this case, it will @b not
   * listen on any sockets.  When this parameter is @c false, Telnet will
   * create other tasks for the shell which listen on sockets.
   */
  bool keep_stdio;
} rtems_tinysshd_config;

/**
 * @brief Start the SSH subsystem with the provided configuration.
 */
rtems_status_code rtems_tinysshd_start(const rtems_tinysshd_config* config);

/**
 * @brief Stop the SSH subsystem with the provided configuration.
 */
rtems_status_code rtems_tinysshd_stop(void);

/**
 * @brief Print the key.
 */
rtems_status_code rtems_tinysshd_printkey(const rtems_tinysshd_config* config);

#ifdef __cplusplus
}
#endif

#endif
